// BASE.CPP
// Implementación de las clases base del controlador y simulador

#include "base.h"
#include <iostream> 

// Clase Dispositivo

int Dispositivo::actualiza() { // Al actualizar, se pide dormir al hilo de ejecución el tiempo	
								  // estipulado en el período de muestreo y que coincide con tiempo_chrono
	std::chrono::high_resolution_clock::time_point fin_compensacion = std::chrono::high_resolution_clock::now();
	std::chrono::duration<long int, std::micro> lapso =  
		std::chrono::duration_cast<std::chrono::microseconds>(fin_compensacion - Dispositivo::inicio_compensacion);	
			// lapso de tiempo calculado. Sirve para la compensación de tiempo	
			
	estados_pausador estado = RUTINA; // Estado de rutina para el pausador	
								
							
	if(Dispositivo::ejecucion && lapso < period_m) { // Si hay ejecución y el lapso no sobrepasa el periodo de muestreo		
													 // el hilo duerme
			std::this_thread::sleep_for(period_m - lapso); 
										// Cuándo hay ejecución, éste duerme un período de muestreo
										// Se resta el tiempo consumido por el ciclo (compensación)
										// Nota: En la máquina virtual no se ha observado la compensación
										// quizá por restricciones/fallos de precisión de la misma máquina virtual
	}
	
	Dispositivo::pausador(estado); // Si se ha solicitado pausa de forma externa, la rutina se bloquea aquí. 
	
	Dispositivo::tiempo++;
	Dispositivo::inicio_compensacion = std::chrono::high_resolution_clock::now(); // Cálculo del tiempo de compensación (inicio)
	return 0;
		
}


Dispositivo::Dispositivo(unsigned int t, int id) { // Constructor que inicializa el valor del muestreo a un valor preestablecido
	
	

	Dispositivo::configura(t,id); // llamada al configurador
    	
}

void Dispositivo::configura(unsigned int t, int id) { // configurador de dispositivo
	
	if(t > 0) { // ¿Es muestreo positivo y por tanto ejecutable? Calculamos período de muestreo
		Dispositivo::muestreo=t;
		Dispositivo::ejecutable=true; // Tenemos frecuencia de muestreo válida y posibilidad de ejecución
		Dispositivo::period = 1/(float)Dispositivo::muestreo; // Calculamos periodo
		Dispositivo::period_m = std::chrono::duration<long int, std::micro>((long int)(Dispositivo::period*1000000)); 
	
	} else {
		Dispositivo::muestreo = 0L; 
	
	} // En caso contrario, poner a cero el tiempo de muestreo y promise de hilo no ejecutable
	

		
	Dispositivo::idslot=id;		
}	
		
unsigned int Dispositivo::obtiene_muestreo() { // Lectura del valor de muestreo del dispositivo
	return Dispositivo::muestreo;
}	
		
int Dispositivo::obtiene_id() { // Lectura del valor de id del dispositivo
	return Dispositivo::idslot;
}	

long unsigned int Dispositivo::obtiene_tiempo() { // Lectura del tiempo que lleva funcionando el dispositivo
	return Dispositivo::tiempo;
	
}
	
float Dispositivo::obtiene_tiempo_f() { // Lectura del tiempo que lleva funcionando el dispositivo en segundos
	return (float)(Dispositivo::tiempo)*Dispositivo::period;
}	

float Dispositivo::obtiene_period() { // Lectura del valor de periodo
	return Dispositivo::period;
}	

bool Dispositivo::obtiene_alarma() { // Obtención de alarma
	return Dispositivo::alarma;
}
                        
void Dispositivo::activa_alarma() { // Activación de alarma

	Dispositivo::alarma=true;
}

bool Dispositivo::obtiene_ejecucion() { // Obtención de condición de ejecución
	return Dispositivo::ejecucion;
}

bool Dispositivo::obtiene_ejecutable() { // ¿Es un dispositivo ejecutable?
	return Dispositivo::ejecutable;
}

bool Dispositivo::obtiene_pausa() { // ¿Está pausado?
	
	return Dispositivo::estado_pausa;

}


int Dispositivo::pausa(bool pausar) { // Pausar/despausar dispositivo. Se utiliza bloqueo por std::mutex pausa.
	
	estados_pausador estado;
	
	if (Dispositivo::estado_pausa == pausar) return -2; // Nada que hacer, ya estaba pausado/despausado, como se quería
		
	if(Dispositivo::ejecucion) {	
	
		if (pausar) { // Se desea pausar			
			estado = BLOQUEO;
			Dispositivo::pausador(estado);
			Dispositivo::estado_pausa = true;
		
		
		} else { // Se desea retomar la ejecución
			estado = DESBLOQUEO; 
			Dispositivo::pausador(estado);
			Dispositivo::estado_pausa = false;
			
		
		}
	} else {
		
		return -1; // Error. No se estaba ejecutando. No se puede pausar lo que no se ejecuta
	}	
	return 0; // Salida normal
}

void Dispositivo::pausador(estados_pausador estado) { // rutina para el bloqueo/desbloqueo (pausador)
	
	switch (estado) {
			case RUTINA: // Zona para la rutina. Si se le requiere pausa, se encuentra con un bloqueo en lock y espera
				Dispositivo::mutex_pausa.lock();
				Dispositivo::mutex_pausa.unlock(); // Se libera el cerrojo. 
				break;
			case BLOQUEO: // Se quiere bloquear la rutina. El gestor (menú) entra en esta zona y deja un lock indefinido
				Dispositivo::mutex_pausa.lock();
				break;
			case DESBLOQUEO: // Se quiere desbloquear la rutina. El gestor libera el cerrojo. 
				Dispositivo::mutex_pausa.unlock();
				break;
	}
	
	return;	
}

int Dispositivo::inicia() { // inicio/continuación de ejecucion de thread. Indispensable que sea ejecutable

	if(Dispositivo::ejecutable && !Dispositivo::estado_pausa && !Dispositivo::ejecucion) { 
																// Es ejecutable, no se está ejecutando
																// y no hay pausa. Se procede.
		if(Dispositivo::hilo.joinable()) {Dispositivo::hilo.join(); // ¿Había hilo aún? Se limpia
		
			std::cout << "\tAVISO: El hilo era joinable"; // No debería llegar a este punto
		}
		// Esta instrucción previene excepciones por errores de programación, como intentar un hilo joinable
		// que, aunque cesó su ejecución según el flag, aún no se unió a la rutina que lo llamó o a la rutina
		// propiedad del hilo. 
		
		
		Dispositivo::ejecucion = true;	// Se informa de que se está ejecutando desde este mismo momento

		Dispositivo::inicio_compensacion = std::chrono::high_resolution_clock::now(); 
										// Cálculo del tiempo de compensación (inicio)
		hilo = this->RutinaThread(); // Se inicia objeto std::thread mediante función lambda (retrollamada a rutina)
		
		return 0; // éxito
	}
	return -1; // En caso de estar ejecutándose, no ser ejecutable o de estar en pausa, devuelve error.
	
} 

std::thread& Dispositivo::acaba(int& error) {// Método que finaliza la ejecución del hilo y la liberación de sus recursos
									// Este método se ha de llamar al acabar el programa o el menú en todos los hilos
	std::thread& hilo = Dispositivo::hilo;		
	
	// Si estamos ante un hilo en ejecución	
	if(Dispositivo::ejecutable && Dispositivo::ejecucion) {		
	
		if(Dispositivo::estado_pausa) // ¿está pausado?
			Dispositivo::pausa(false); // Se solicita que el hilo retome la ejecución
			
		Dispositivo::ejecucion = false; // Finaliza el ciclo y se pide el cierre al hilo por flag
									    // en rutina() deja de cumplirse la condición de bucle	
	
										 
		error = 0; // Ya se envió la consigna	
	
	} else 
			error = -1; // En caso contrario, devuelve error	(no era ejecutable ni se estaba ejecutando)
	        
	
	return hilo;
} // Nota: Se devuelve el código de error por referencia en el argumento. Este sirve para llamar o no a
  // std::thread.join(), puesto que se puede haber parado el hilo previamente. La rutina principal ha de hacer la
  // comprobación con el código error de retorno. También se dispone de std::future para avisar al hilo llamante de 
  // que acabó la ejecución y no hacerle esperar con el método join. 

  
int Dispositivo::desactiva_ejecucion(){ // Desactiva ejecución del hilo
	if(Dispositivo::ejecutable && !Dispositivo::ejecucion) 
		Dispositivo::ejecutable = false; // Desactivación efectiva si es ejecutable y no se estaba ejecutando
	else return -1; // Error si no es así 
	
	return 0; // Correcto

}	

void Dispositivo::rutina() { // Rutina de dispositivo. Genérica. No hace nada en esta clase, pero sirve de modelo. 
							 // Conviene estructurarlo así en las clases derivadas, para que de esta
							 // forma finalice su ejecución al dejar de ser ejecutable 

	
	while (Dispositivo::obtiene_ejecucion()) {		
		Dispositivo::actualiza();		
	}
	
	return; 

} 
	

void Dispositivo::propiedades() { //Propiedades en pantalla
	
	std::string ejecutable = Dispositivo::obtiene_ejecutable() ? "SI" : "NO";
	std::string ejecucion = Dispositivo::obtiene_ejecucion() ? "SI" : "NO";
	std::string pausa = Dispositivo::obtiene_pausa() ? "SI" : "NO";
	std::string alarma = Dispositivo::obtiene_alarma() ? "SI" : "NO";	
	
	std::cout << "\nPROPIEDADES DEL DISPOSITIVO:\n "; 
	std::cout << "\n DISPOSITIVO ID = " << Dispositivo::obtiene_id() << " | Freq muestreo: " 
	<< Dispositivo::obtiene_muestreo() << " Hz | Periodo: " << Dispositivo::obtiene_period() << " seg.";
	std::cout << "\n Tiempo ejecución: " << Dispositivo::obtiene_tiempo_f() << " seg. "; 
	std::cout << "\nEjecutable: " << ejecutable << " | Ejecución: " << ejecucion << " | Pausa: " << pausa <<
				 " | Alarma: " << alarma << "\n";
	
	return;	
}

void Dispositivo::propiedades_base() { //Propiedades de clase base. Se indica la misma que Dispositivo
	
	Dispositivo::propiedades();
	
	return;
}



Dispositivo::~Dispositivo() { // Destructor base de Dispositivos. Nada que liberar  

}

template <typename T>
Dinamico<T>::Dinamico(unsigned int t, int id, const std::vector<unsigned int>& v) : Conectable(t,id) { // Constructor

	Dinamico<T>::configura(v);
		
}

template <typename T>
Dinamico<T>::Dinamico(unsigned int t, int id) : Conectable(t,id) { // Constructor con sólo t e id (sin iniciar colas)
	
		
}

template <typename T>
void Dinamico<T>::configura(const std::vector<unsigned int>& v) { // configurador del vector de colas

	unsigned int i;
	

	if (v.size()==0) return; // ¿vector nulo? no se admite
	Dinamico<T>::estado.resize(v.size()); // Establecemos vectores de estado y siguiente
	Dinamico<T>::siguiente.resize(v.size());	
	for (i=0;i<v.size();i++) 
		Dinamico<T>::estado[i].assign(v[i],0); // Establece longitud de cada cola, rellena de ceros		
		
}

template <typename T>
int Dinamico<T>::actualiza() {	// Actualización de colas en el sistema dinámico
	
	unsigned int i; 
	int resultado=0;
	if (Dinamico<T>::estado.size()!= Dinamico<T>::siguiente.size()) resultado++; 
	// Si no coinciden longitudes de vector, error. 
	else {
		for (i=0;i<estado.size();i++) {
			if(estado[i].size()==1) estado[i][0]=siguiente[i];
			// Para colas de un elemento no hace falta hacer push-pop sino sustituir directamente el elemento
			else {		
				estado[i].push_front(siguiente[i]); //siguiente pasa a ser el elemento cero de la cola
				estado[i].pop_back(); //el último elemento de la cola desaparece
				// El tamaño de cola permanece invariante
			}
		}
	
		
	}
	resultado += Conectable::actualiza(); // Actualizar clase base
	return resultado;
	
	
}

// Propiedades dinamico

template <typename T>
void Dinamico<T>::propiedades() { // Propiedades de Dinamico	
	
	std::cout << "\nPROPIEDADES DINAMICO:\n "; 
	std::cout << "Longitud vector de colas : " << Dinamico::estado.size() << "\n";
	
	if(Dinamico::estado.size() > 0) { // Si hay vector de colas, se imprime longitud de cada cola
		std::cout << "Longitud de cada cola: {";
		for(unsigned int i = 0 ; i < Dinamico::estado.size(); i++) {
			if(Dinamico::estado.size() - i > 1) // No es el último elemento (añadir coma)
				std::cout << Dinamico::estado[i].size() << ", "; 
			else
				std::cout << Dinamico::estado[i].size(); // Ultimo elemento (sin coma añadida) 
		}
		std::cout << "}\n";
		
	}	
	
	return;		
}

// Propiedades base

template <typename T>
void Dinamico<T>::propiedades_base() { //Propiedades base. Siguiente: Conectable (superclase inmediatamente superior)
	
	Dinamico::propiedades();
	
	Conectable::propiedades_base();
	
	return;
}

// Clase Continuo (sistema dinámico continuo)

template <typename T>
Continuo<T>::Continuo(unsigned int t, int id, const std::vector< std::vector<T> >& a, const std::vector<T>& b) 
							: Dinamico<T>(t,id) {
		Continuo<T>::configura(a, b); // llamada al configurador con vectores
		
		
}

template <typename T> 	// Constructor mínimo. Se crea una sola cola de un elemento
Continuo<T>::Continuo(unsigned int t, int id) : Dinamico<T>(t,id){ 

	
						
	Continuo<T>::configura(); // llamada al configurador básico
}

template <typename T> // configurador con inicialización de vectores
void Continuo<T>::configura(const std::vector< std::vector<T> >& a, const std::vector<T>& b) {
	unsigned int n;
	n = b.size();
	if (n >0 && n != a.size()) return; // la matriz ha de coincidir en filas con el vector de términos independientes
								         // no ha de ser de longitud nula
										 // ///////// Comprobar vector a vector
	std::vector<unsigned int> ve (n,1);	// Las colas son de un elemento (aplicaremos métodos de un solo paso)
	Dinamico<T>::configura(ve); // inicializa base (vector dimensión n de colas de un solo elemento)
	Continuo<T>::asoc = a; // Copiar matriz asociada
	Continuo<T>::indep = b; // Copiar vector de términos independientes
	Continuo<T>::orden=n;
	Continuo<T>::eul.resize(n); // Inicializa auxiliares
	Continuo<T>::k1.resize(n);
	Continuo<T>::k2.resize(n);
	Continuo<T>::k3.resize(n);
	Continuo<T>::k4.resize(n);
	
}

template <typename T>
void Continuo<T>::configura() { // configurador sin inicialización de vectores
	std::vector<unsigned int> ve (1,1);	
	Continuo<T>::orden=1;
	
	Continuo<T>::asoc.resize(1); // Revisar /////////////
	Continuo<T>::asoc[0].resize(1);
	Continuo<T>::indep.resize(1);
	
	Continuo<T>::eul.resize(1); // Inicializa auxiliares
	Continuo<T>::k1.resize(1);
	Continuo<T>::k2.resize(1);
	Continuo<T>::k3.resize(1);
	Continuo<T>::k4.resize(1);
	Dinamico<T>::configura(ve); // configura clase base dinamico
}

template <typename T>
int Continuo<T>::actualiza() { // Actualizador de clase, se limita a llamar a los actualizadores base
	return Dinamico<T>::actualiza(); 
}

template <typename T>
void Continuo<T>::eval(std::vector<T>& entrada, std::vector<T>& salida, T propor) { // Función auxiliar para RK4
	unsigned int i,j;   
		// Vector salida = matriz asociada * (vector estados + (vector entrada * propor))+vector t. independientes
	T aux;	
		
	for(i=0; i < Continuo<T>::orden; i++) { 
		aux=0;		
		for(j=0; j < Continuo<T>::orden; j++){
			aux+=Continuo<T>::asoc[i][j]*(Dinamico<T>::estado[j][0]+(propor*entrada[j]));
		}
		salida[i]=aux+Continuo<T>::indep[i]; 
	}
}

template <typename T>
void Continuo<T>::eval(std::vector<T>& salida) { // Función auxiliar para Euler y RK4
	unsigned int i,j;   // Vector salida = matriz asociada * vector estados + vector t. independientes
	T aux;	
		
	for(i=0; i < Continuo<T>::orden ; i++) { 
		aux=0;		
		for(j=0; j < Continuo<T>::orden ; j++){
			aux+=Continuo<T>::asoc[i][j]*Dinamico<T>::estado[j][0];
		}
		salida[i]=aux+Continuo<T>::indep[i]; 
	}
}

template <typename T>
void Continuo<T>::euler() { // Método de Euler, sistemas de ecuaciones diferenciales
	unsigned int i;
	T p;

	p=(T)Dispositivo::obtiene_period(); // Obtener paso (período de muestreo)
	
	Continuo<T>::eval(Continuo<T>::eul); // evaluación del sistema
	
	
	
	for(i=0; i < Continuo<T>::orden ; i++)  
		// el siguiente término es el anterior más lo evaluado multiplicado por el período
		Dinamico<T>::siguiente[i] = Dinamico<T>::estado[i][0] + p*Continuo<T>::eul[i]; 	
	
}

template <typename T>  
void Continuo<T>::rk4() { // Método de Runge-Kutta orden 4, sistemas de ecuaciones diferenciales
	unsigned int i;
	T p;
	
	p=Dispositivo::obtiene_period(); // Obtener paso
	
	Continuo<T>::eval(Continuo<T>::k1); // obtención de k1 (evaluación de sistema)
	Continuo<T>::eval(Continuo<T>::k1,Continuo<T>::k2,p/2); // obtencion de k2
	Continuo<T>::eval(Continuo<T>::k2,Continuo<T>::k3,p/2); // obtencion de k3
	Continuo<T>::eval(Continuo<T>::k3,Continuo<T>::k4,p); // obtencion de k4
	
	for(i=0; i < Continuo<T>::orden ; i++)  // determinación del vector siguiente
		Dinamico<T>::siguiente[i] = Dinamico<T>::estado[i][0] + 
		((p/6)*(Continuo<T>::k1[i]+2*Continuo<T>::k2[i]+2*Continuo<T>::k3[i]+Continuo<T>::k4[i]));	
	
}

// Propiedades Continuo

template <typename T>
void Continuo<T>::propiedades() { // Propiedades de Continuo
	
	std::cout << "\nPROPIEDADES CONTINUO:\n "; 
	std::cout << "Dimensiones matriz asociada : " << Continuo::asoc.size() << " x " <<
		Continuo::asoc.size() << "\n";
	std::cout << "Dimensiones vector términos independientes : " << Continuo::indep.size() << "\n";
	
	return;
}

// Propiedades base

template <typename T>
void Continuo<T>::propiedades_base() { //Propiedades base. Siguiente: Dinamico
	
	Continuo<T>::propiedades();
	
	Dinamico<T>::propiedades_base();	
	
	return;
}


template class Dinamico<unsigned int>; // Plantillas que se usarán para los dispositivos. Se han de declarar aquí.
template class Dinamico<float>;
template class Continuo<float>;        // De esta forma se enlazará adecuadamente














