// BASE.H
// Declaraciones de todas las clases. Jerarquía de clases en la memoria adjunta
// Se parte de una clase base llamada Dispositivo que otorga temporización, posibilidad de ejecución de hilos, alarmas, etc.
// Esta clase base conforma los dispositivos dinámicos, los generadores de funciones (no dinámicos), los conectores
// y los watchdog

// A partir de la clase de objetos dinámicos distinguimos los continuos (para planta) y los discretos (para controladores)

// En la clase de conectores tenemos los conectores estándar para conectar diferentes dispositivos de planta 
// y una subclase sumadora/comparadora para los lazos de control que nos permitirá conectar controladores 
// formando sistemas complejos, configuraciones en cascada, etc. 

#ifndef BASE_H
#define BASE_H
#include <vector>
#include <queue>
#include <deque>
#include <thread>
#include <mutex>
#include <atomic>		
#include <chrono>
#include <future>
#include <cstring> // necesario para std::memset (relleno de zonas de memoria)

	#define BYTE unsigned char // Definición de byte como unsigned char (para buffers dinámicos)
	
	#define CODIGO_FIN_HILO 965598

	enum tipo_proceso {PROCESO_CONTROLADOR, PROCESO_PLANTA};
	
	enum estado_bus {BUS_NORMAL, BUS_PARADA};	
	
	enum tipo_datos_rec {FLOAT, DOUBLE, INT, CHAR, BOOL}; // Tipo de datos a recopilar por conector_rec y grabador
	
	struct cabecera_sv_conteo { // Cabecera de los datos recibidos en cola de mensajes. mtype es el id del discretizador
		long mtype;        // y conteo_cola da información del número de elementos que quedan para ser recibidos o
		int conteo_cola;   // enviados. Si alcanza un número negativo es que el bus ha decidido terminar de transmitir
		enum estado_bus estado; // Estado del bus NORMAL o PARADA (en caso de parada se informa al otro bus)
	};
	
	#define HUELLA1 0xA179CB8A // Huellas para los datos de cabecera en mensajería. (Se pueden escoger libremente)
	#define HUELLA2 0x4F216B11
	
	struct plantilla_datos_discretizador_sv { // Características importantes del discretizador transmitible por bus		
		unsigned long huella1; // Código de huella, identifica la cabecera y sus datos como válidos
		unsigned long huella2;
		int id; // id de dispositivo
		int esid; // esid de conexion con el dispositivo de planta o controlador
		unsigned int tamano_cola; // Tamaño de cola
		unsigned int fmuestreo; // Frecuencia de muestreo (han de coincidir a ambos lados)
		unsigned int lal_dispositivo; // Tamaño de datos al dispositivo desde el bus.
		unsigned int lal_bus; // Tamaño de datos al bus, desde el dispositivo.
		bool recibido; // "Recibí la cabecera. Puedes empezar a transmitir datos"
	};	
	
	class Conector_rec; // Forward Declaration
	
	struct plantilla_datos_rec {
			int id; 	// id del conector grabador
			Conector_rec *conector; // Puntero al conector (para delegar la llamada de recolección de datos al grabador)			
			unsigned int pos; // Posición del dato a recopilar
			tipo_datos_rec tipo; // Tipo del dato a recopilar (nos proporciona su longitud con sizeof() )
			bool io; // false = entrada; true = salida. Selector de dirección de flujo de datos respecto primera conexion
			BYTE *dir = nullptr; // Dirección exacta (en el buffer rápido de Conector_rec)
			std::string etiqueta; // Etiqueta string del dato (para mostrarse en el volcado .dat)			
	};

class Dispositivo { // Clase base de un dispositivo en tiempo discreto (planta o controlador)
	unsigned int muestreo; // frecuencia de muestreo: 1 = un muestreo cada segundo; 2 = dos muestreos cada segundo, etc. 
	std::atomic<long unsigned int> tiempo {0}; // Tiempo que lleva funcionando el dispositivo en unidades de muestreo	
	int idslot=0; // Identificador de dispositivo - Nota: No se puede asignar un valor nulo (0 o neg) en discretizadores
										// ni tampoco los discretizadores pueden tener id > 99999
	float period=0.; // Periodo, inverso de la frecuencia de muestreo. Usado en el paso p de métodos numéricos p.ej.
	
	std::chrono::duration<long int, std::micro> period_m; // Periodo en formato chrono (microsegundos) 
	
	//Nota: Las variables std::atomic o atómicas permiten sincronizar la lectura y escritura de los datos,
	// de tal forma que dos hilos que convergan en esta variable no pueden leer/escribir a la vez, de esta
	// forma se garantiza que no haya corrupción de datos o toma de decisiones erróneas que podrían provocar estados
	// indeterminados e imprevisibles de ejecución. 
	std::atomic<bool> alarma {false};// Alarma del dispositivo. Si alarma=true significa que hay un problema
	std::atomic<bool> ejecutable {false};// ¿Es ejecutable? Constatarlo al construir el dispositivo (muestreo >= 0)
	std::atomic<bool> ejecucion {false};// ¿Se está ejecutando un hilo?
	std::atomic<bool> estado_pausa {false};// Flag de pausa, se pone a true si al dispositivo se le indicó que se pausase
	
	std::mutex mutex_pausa; // Mutex de pausa de thread. Se usa en acciones de pausa/retomar, por selección del usuario.
	
	
	enum estados_pausador {RUTINA, BLOQUEO, DESBLOQUEO};
	
	void pausador(estados_pausador estado);
	
	// Nota: El dispositivo será ejecutable si hay una frecuencia de muestreo mayor que cero. El muestreo 
	// es nulo para los dispositivos pasivos como conectores y bus pues no son ejecutables.
							
	std::thread hilo; // Hilo (thread) del dispositivo. Cada dispositivo ejecutará su thread. Permanece 
					  // interno (privado o protected si se estimara oportuno) y sólo se devolverá en forma
					  // de referencia al acabar. De esta forma el proceso principal puede liberar los recursos del
					  // hilo con std::thread.join();
	
	std::thread RutinaThread() {
          return std::thread([=] { rutina(); }); // Función lambda (retrollamada) que suministra la
									// dirección de la rutina de ejecución al constructor de std::thread	
    }
	
	 std::chrono::high_resolution_clock::time_point inicio_compensacion; 	
													// Marca de tiempo de compensación. Reduce el valor de sleep
													// si fuese necesario (SU EFECTIVIDAD DEPENDE DE LA RESOLUCION 
													// DEL RELOJ DEL SISTEMA, DE LA MAQUINA Y DE LAS FREQ. MUESTREOS)
protected:		

	

	
	std::mutex exclusion; // exclusión mutua de los hilos que concurren en un objeto de esta clase **Necesario**
						  // Este mutex se utilizará en métodos generales, preferentemente consulta, entrada/salida 
						  // puntual en un objeto de la clase Dispositivo, sin que participen threads que necesitan
						  // muchos recursos de procesador (simulación, por ejemplo. En este último caso 
						  // Hay otros mutex más especializados y que están ubicados en lugares de
						  // frecuente intercambio de datos como por ejemplo conectores y bus. Es recomendable que cada
						  // hilo acapare lo menos posible el lapso lock-unlock durante la exclusión mutua y garantizar
						  // que la exclusión queda desbloqueada, incluso si hay error (o una excepción).
						  
	 

//  Clase clase; // Tipo de dispositivo

	virtual int actualiza(); // Actualización de los valores y estados de los dispositivos 
	                         // En esta clase, actualización del tiempo (incremento del tiempo en uno)  
	virtual void rutina(); // Rutina del thread de dispositivo. Polimórfico. Depende del tipo de dispositivo
	void configura(unsigned int t, int id); // Inicializa variables del dispositivo
	void activa_alarma(); // activar alarma en el dispositivo, internamente, por alguna de las subclases
	int desactiva_ejecucion(); // Desactiva más ejecuciones del dispositivo (-1 si se está ejecutando o no ejecutable)
	
	

public: // métodos públicos
	Dispositivo(unsigned int t, int id); // Constructor (inicializa muestreo e identificación slot)	 
	unsigned int obtiene_muestreo(); // Lectura del valor de muestreo del dispositivo
	int obtiene_id(); // Obtener id del dispositivo
	long unsigned int obtiene_tiempo(); // Lectura del tiempo que lleva funcionando el dispositivo
	float obtiene_tiempo_f(); // Lectura del tiempo que lleva funcionando el dispositivo en segundos
	float obtiene_period(); // Lectura del período de muestreo del dispositivo
	
	int inicia(); // iniciar rutina del dispositivo (si no es ejecutable, se devuelve error -1)
	int pausa(bool pausar); // Pausar dispositivo	
	std::thread& acaba(int& error); //Acabar rutina del dispositivo 
	 // Devuelve referencia al thread para realizar el join posterior y también un error que se devuelve por referencia
	 // en el argumento. Si error = -1 no es necesario que el proceso principal realice std::thread.join()
							
	virtual void propiedades(); // Método polimórfico (dependiente del objeto) que muestra sus propiedades en pantalla
	virtual void propiedades_base(); // Método polimórfico de las propiedades base
	
	bool obtiene_alarma(); // ¿hay alarma en el dispositivo?
	bool obtiene_ejecucion(); // ¿El dispositivo se está ejecutando?
	bool obtiene_ejecutable(); // ¿Es ejecutable? 
	bool obtiene_pausa(); // ¿Está pausado? (detenido de forma indeterminada hasta que se decida retomar)	

	virtual ~Dispositivo(); // Destructor virtual. Es necesario cuando borramos (delete) un puntero de la clase base
							// puesto que hay que llamar a los destructores de las clases derivadas. Como en el
							// menú interactivo sólo tenemos los punteros a la clase base (Dispositivo *) se hace
							// necesario liberar los objetos mediante destructores virtuales. 
};

class Conectable; // Necesaria esta declaración, para ser reconocida por Conector (Forward Declaration)

class Conector : public Dispositivo { //conector de dispositivos y exclusión mutua para que éstos puedan
									  // compartir datos en tiempo real sin comprometer la integridad de los datos	 
protected: 

	struct plantilla_conector {
		Conectable *conectables[2]={nullptr, nullptr}; // Punteros a los conectables enlazados
		int id[2] = {0,0}; // ID de los dispositivos
		int esid[2] = {0,0};  // Identificadores de conexión, longitudes de entradas y salidas
		unsigned int lentradas[2] = {0,0};
		unsigned int lsalidas[2] = {0,0};
		
		
		std::atomic<unsigned long int> accesos[2]; // Número de accesos por cada flujo de datos
										  // Si se multiplica por la longitud de paquete nos da el flujo de datos
										  // en bytes. 	
		std::atomic<bool> escrito_leido[2]; // leido_escrito es un flag con 0=dato leido y 1=dato escrito que indica
										// a quien lee o escribe si en el otro lado actualizaron información o leyeron
										// Con esto, evitamos escribir o leer innecesariamente y se ahorra tiempo de
										// proceso / CPU, principalmente cuando ambos procesos no operan a la misma
										// frecuencia o uno de los dos se para/espera respuesta de otro dispositivo. 	
		BYTE *buffers[2] = {nullptr, nullptr}; // Buffers de entrada/salida. Se hace necesario recurrir a esta 
										// memoria intermedia dentro de la zona de exclusión mutua para garantizar
										// que cada dispositivo lee los datos íntegros y no mientras están siendo
										// actualizados por el otro dispositivo, produciéndose una condición de
										// carrera y una corrupción de los datos leídos. 
		BYTE *entrada[2] = {nullptr, nullptr}; // Entrada de datos al conector	
		BYTE *salida[2] = {nullptr, nullptr}; // Salida de datos al conector	
	};	
	
	plantilla_conector conector; // Estructura de datos del conector. 
	
	std::atomic<long int> conteo_accesos; // Número de accesos al conector para E/S


	int actualiza(); // Actualización (innecesaria en esta clase)
	
	std::mutex exclusion_entradasalida; // exclusión mutua de los hilos que concurren en el método entradasalida
public: 

	Conector(int id); // Constructor preferente del conector. Sólo su id. No es ejecutable ni tiene thread interno. 
	~Conector(); // Destructor que libera los buffers asignados cuando el objeto se libera de la memoria.

	/* bool primera_actualizacion(Conectable *conectable); // Verificar que ya existe el primer dato procedente
																  // del otro dispositivo conectado */
	void propiedades(); // Propiedades del conector
	void propiedades_base(); // Propiedades de base (aquí será la de dispositivo)

	virtual int registrar(Conectable *disp, int esid, BYTE *entrada, BYTE *salida,
								unsigned int lentrada, unsigned int lsalida); // Función polimórfica 
					// Registro de dispositivo en el conector. 
			// Se suministran puertas de entrada, salida, longitude de paquete, así como identificador e/s.	
	int datos_complementario(int *esid, unsigned int *lentrada, unsigned int *lsalida);
					// Datos del complementario para un dispositivo discretizador que enviará los datos a un bus. 
					// Se le proporciona los datos que le complementan con el dispositivo al otro lado del conector. 
					
	int entradasalida(Conectable *conectable, bool solo_lectura); 
						// Se efectúa la actualización E/S del dispositivo
						// Es conveniente, en dispositivos intermedios como discretizadores 
						// no escribir hasta tener un dato válido, por eso pueden habilitar el flag solo_lectura
						// en las condiciones necesarias. Así se garantiza que se entregan los datos correctos
						// los cuales son responsabilidad del dispositivo que envía. 
	
	
};

class Grabador; // Forward declaration

class Conector_rec : public Conector { // Conector con un método de recolección de datos. 
									   // Será utilizado por el grabador. 

	
	BYTE *buffer_sondeo_entrada = nullptr; 
	BYTE *buffer_sondeo_salida = nullptr; // Buffers propios. Se copian de los buffers del conector por memcpy (rápido) 
								  // procurando, así, el menor impacto posible en el funcionamiento normal del conector
								  // durante el bloqueo por mutex
								 
	std::vector<plantilla_datos_rec> datos; // Datos a ser recopilados cuando se inicie el grabador
	
protected:
	std::mutex exclusion_vector; // Exclusión mutua en el manejo de vectores 
	unsigned int obtiene_sizeof(tipo_datos_rec tipo); // Obtener sizeof() de un tipo por argumento (tipo_datos_rec)
	
public:
	Conector_rec(int id, Grabador *grabador); // Constructor
	~Conector_rec(); // Destructor (Importante: si se usa el menú ha de ser un destructor virtual)
	
	int registrar(Conectable *disp, int esid, BYTE *entrada, BYTE *salida,
								unsigned int lentrada, unsigned int lsalida); // Modo propio de registro
														// Hereda el registro de conector y le añade algunas 
														// tareas como crear los buffers de recolección
	int crea_sonda(bool io, unsigned int pos, tipo_datos_rec tipo, std::string etiqueta); 
														// Crear sonda para recoger datos
	
	
	
	const std::vector<plantilla_datos_rec>& obtener_datos(); // Obtener los datos del conector. 
	
	void recoleccion(); // Recolección de datos (llamado por el grabador). 
	
	void propiedades(); // Propiedades del conector
	void propiedades_base(); // Propiedades de base (Dispositivo)
		
};


class Grabador : public Dispositivo { // Clase grabador
	
	// Datos a ser guardados y posteriormente volcados. Son matrices de datos, puesto que se puede
	// volcar más de un dato por tipo, al hacer recolección de todos los conectores.
	
	std::vector<Conector_rec *> conectores; // lista de conectores
		
	tipo_proceso tipo; // Tipo de proceso (planta o controlador) sirve para etiquetar los archivos de salida
		
	std::vector<std::queue<float>> datos_float;	// Vectores de colas
	std::vector<std::queue<double>> datos_double;
	std::vector<std::queue<int>> datos_int;
	std::vector<std::queue<char>> datos_char;
	std::vector<std::queue<bool>> datos_bool;
		
	std::vector<plantilla_datos_rec> datos; // Datos a ser recopilados, de todos los conectores_rec asociados
											// Sólo actualizará este vector si, a petición de uno de los 
											// conectores, el grabador está detenido. Si estuviera en marcha
											// se corromperían los datos ya guardados. Hay que esperar a que 
											// se detenga completamente (no es suficiente pausarlo). 
	
	
	std::atomic<unsigned int> ifloat {0}; // Esta forma de iniciar objetos atomic (los datos atómicos son objetos)
	std::atomic<unsigned int> idouble {0}; // es compatible con versiones iguales o superiores a C++11
	std::atomic<unsigned int> iint {0};
	std::atomic<unsigned int> ichar {0};
	std::atomic<unsigned int> ibool {0};
	
	std::atomic<unsigned long int> cuenta; // Cuenta de todas las muestras grabadas. Se resetean en cada reinicio de 
											// hilo (detención-inicio). La pausa no resetea esta cuenta
											
	std::atomic<unsigned int> uso_bytes {0}; // Uso en kilobytes
	std::atomic<bool> limite; // Se llegó al límite admisible y hay que dejar de almacenar/contear
											
	std::atomic<unsigned int> cuenta_global {0}; // Numero de detenciones e inicios (para el nombre del archivo)
	std::atomic<bool> diferentes {true}; // ¿Diferentes nombres de archivo en una misma sesión?
	std::mutex mutex_grabador; // Mutex del grabador
	
	std::atomic<unsigned int> kbs; // Límite del almacenamiento en kilobytes. Se evita desbordamiento. 
									 // Si kbs=0 el almacenamiento se considera ilimitado (lo que podría agotar la memoria
									 // del computador si no se vigila bien y se deja funcionando largo tiempo). 
	
	void actualizar_datos(); // actualizador de datos
	
	void archivo(); // Guardar resultados a archivo, una vez que finaliza la rutina (detención del hilo). 
		
protected:
	void rutina(); // Rutina del grabador. No es la rutina genérica. Necesita realizar operaciones finales
						// tras el volcado de datos, como guardar a archivo
	int actualiza(); // Actualizador
public:
	Grabador(unsigned int t, int id, tipo_proceso tipo, bool diferentes, unsigned int kbs); // Constructor
	
	int registrar(Conector_rec *conector); // Registra un conector_rec
	
	void propiedades(); // Propiedades del grabador
	void propiedades_base(); //Propiedades de clase base. Se indica la misma que Dispositivo
	
};

template <typename T>
class Controlador; // Necesario declararlo para Watchdog (Forward Declaration)

template <typename T>
class Watchdog : public Dispositivo { // Clase watchdog, derivada de dispositivo y asociada a controladores

	enum estado_detencion_plantilla {CONTROLADOR_NORMAL, CONTROLADOR_DETENIENDOSE, CONTROLADOR_DETENIDO};
	
	std::mutex exclusion_watchdog; // Exclusión mutua por ser comprobacion(*Controlador controlador) una rutina
								   // que puede ser accedida de forma recurrente
	
	struct plantilla_controlador {
			Controlador<T> *controlador; // Controlador asociado			
			int id; //id y fmuestreo del controlador 
			unsigned int fmuestreo; 
			unsigned int multiplo; // Múltiplo o conteo máximo admisible
			unsigned int conteo; // Conteo 
			bool detencion; // ¿Necesita el dispositivo a controlar una detención en caso de alarma? 
			estado_detencion_plantilla estado_detencion; // estado de la detencion
							   // eso significa que ha de revisar el estado de ejecucion del controlador y aplicar join()
			std::shared_future<unsigned long int> estado_ejecucion; 
								
								// se utilizará para evitar que el watchdog tenga que esperar al ejecutar join()
								// ya que ejecutará este comando cuando el controlador confirma que sale de su rutina
//			std::thread hilo = std::thread(); // hilo de ejecución
			int err; // Retorno en la detención del hilo
			bool alarma; // Alarma, visible en propiedades
		
	};
	
	struct plantilla_detencion { // Estructura de hilos pendiente de detención, con el thread transferido std::move
		std::thread hilo;
		std::future<long unsigned int> futuro; // Estado del controlador, por mensaje asíncrono (future) 
		unsigned int indice;
	};
	
	std::vector<plantilla_detencion> hilos_pendientes; // Hilos pendientes de detencion
	
	 // multiplo: número de ciclos completados del watchdog que activarían la alarma en caso de que
	 // el controlador no hiciera llamada de comprobación desde la última vez
	 // conteo: conteo de ciclos de watchdog desde la última actualización del controlador
	 // si este conteo sobrepasa al valor preasignado (multiplo) se activa la alarma
	
	std::vector<plantilla_controlador> controladores; // vector de controladores inscritos en el watchdog
	
	int futuro_preparado(std::future<long unsigned int> const& futuro); // Verifica si está el "futuro" preparado
	 
protected:

	int actualiza(); // Actualización del watchdog, realizado por el thread del mismo cada período de muestreo
	void rutina(); // Rutina del thread
	
							 
public:
	Watchdog(unsigned int t, int id); // Constructor del watchdog
	
	int registrar(Controlador<T> *controlador, unsigned int multiplo, bool detencion); // Registro de controlador
																				// Conteo máximo antes de la alarma
																				// Y posibilidad de detencion 
																				// del controlador
	void sigo_operativo(Controlador<T> *controlador); 
							// Llamada de comprobación hecha desde el controlador y que resetea el conteo
							
	void propiedades(); // Propiedades del watchdog
	void propiedades_base(); // Propiedades de base (Dispositivo)
	
	
};

class Conectable : public Dispositivo { // Dispositivos conectables.
										// permite una comunicación entre dispositivos mediante clases "conector"
	unsigned int num_elementos = 0; // número de elementos 	

	struct parametros_conectable {
		Conector *conector;  // Conector asociado al dispositivo
		unsigned int puerto; // Puerto del dispositivo
		bool puerto_asignado; // ¿está el puerto asignado a un conector?
		int esid; // identificación de entrada/salida
		BYTE *entrada; // Puntero de la entrada
		BYTE *salida; // Puntero de la salida
		unsigned int lentrada;  // tamaño de la entrada
		unsigned int lsalida;  // tamaño de la salida
		bool solo_lectura; // Este puerto es de solo lectura. El dispositivo aún no puede ofrecer datos válidos.
		int err; // Información numérica de E/S (si se pudo leer y/o escribir).
	};
	
	std::vector<parametros_conectable> parametros;
	

	std::atomic<int> err_reciente {0}; // err más reciente. Es manejable cuando se tiene un puerto. 
									   // Por ejemplo, en discretizadores. 
									   
					  // Nota: Estos son los códigos de retorno de los conectores (que también se usará en E/S de bus)
					  // err = 1 solo se ha podido escribir en uno u otro lado (entrada respecto al conector) 
					  // err = 2 solo se ha podido leer desde uno u otro lado (salida respecto al conector)
					  // err = 3 se ha podido escribir y leer	
																		
	std::mutex mutex_conectable; // Exclusión mutua propia de la clase						  
protected:

	bool comunicable = true; // Por defecto, nos podemos comunicar. Si así no quisiera la clase derivada, 
							 // lo modifica a false y no intercambian los puertos en cada interacción. 
	
/*	int leer_err(unsigned int i); // Leer estado de la última entrada/salida en el puerto i */
//	void establecer_comunicable(bool flag); // ¿Queremos hacer el dispositivo comunicable? 
	int modificar_solo_lectura(unsigned int puerto, bool solo_lectura); // Modificar el flag de solo lectura en un puerto
	int actualiza(); // Actualización en las comunicaciones y conexiones
	int nuevo_puerto(unsigned int puerto, int esid, BYTE *entrada, BYTE *salida,
							unsigned int lentrada, unsigned int lsalida); 
						// Añade un nuevo puerto de comunicaciones que se puede conectar 
						// Este método se suele llamar en el momento de construir la clase del dispositivo conectable
						// El puerto queda disponible y con el método "conectar" se le puede dar uso si queremos
	int obtiene_err_reciente(); // Obtener el error reciente (para la confirmación lectura/escritura)
	
public:

	Conectable(unsigned int t, int id);
	int conectar(Conector *conector, unsigned int puerto); // Conectar un puerto del dispositivo a un conector	
	
	void propiedades(); // Propiedades del conectable
	void propiedades_base(); // Propiedades base del conectable (Dispositivo)

};

template <typename T>
class Dinamico : public Conectable { // Dotamos al dispositivo conectable de las propiedades de los sistemas dinámicos
									  // En esencia, tiene capacidad de memoria y conserva estados anteriores
									  // Esta clase sirve de base para dispositivos de planta (físicos, continuos) y 
									  // controladores (discretos)
protected:	
	std::vector< std::deque<T> > estado; // Vector de colas con todos los estados
	
	std::vector<T> siguiente; // Valor siguiente que se asigna para cada cola 
	int actualiza(); // Inserta nuevos valores de cola y actualiza clase base
	void configura(const std::vector<unsigned int>& v); //inicializa variables de dinamico 
														// y superclase (dispositivo)
	
public:
	Dinamico(unsigned int t, int id, const std::vector<unsigned int>& v); 
			// El constructor necesita frecuencia de muestreo, id y vector de longitudes de cola
	Dinamico(unsigned int t, int id); // Constructor sin inicializar colas
	void propiedades(); // Propiedades de Dinamico
	void propiedades_base(); // Propiedades base 

	
};

template <typename T>
class Continuo : public Dinamico<T> { // Dispositivos dinámicos continuos. 
								   // Sistemas de ecuaciones diferenciales de primer orden.
									// Implementación preferente para dispositivos físicos (planta)
								   
	std::vector<T> eul; // Vector utilizado en la función Euler 
	std::vector<T> k1; // Vector generado en la función rk4
	std::vector<T> k2; // Se utilizan aquí y no como vectores locales del método. Son definidos en los constructores
	std::vector<T> k3; // De esta forma no se inicializan en cada llamada a los métodos de cálculo 
	std::vector<T> k4; // Ahorrando tiempo de cómputo
	unsigned int orden = 0; // Orden del sistema (matriz cuadrada)
	void eval(std::vector<T>& entrada, std::vector<T>& salida, T propor); // Función auxiliar de evaluación
	void eval(std::vector<T>& salida);
		// Esta función auxiliar evalúa la matriz asociada con el término anterior más 
		// la variable de entrada multiplicada por la constante proporcional propor 
		// más el vector de términos independientes
		// obteniéndose la salida. Utilizada por el algoritmo de Runge-Kutta. 
		// La sobrecarga simplemente multiplica la matriz asociada con el estado anterior más el vector t. indep.  
	
protected:							   
	std::vector< std::vector<T> > asoc; // Matriz asociada a un sistema diferencial lineal cualquiera
	std::vector<T> indep; // Vector de términos independientes del sistema diferencial lineal

	// Resolución de un solo paso (Euler y RK-4) los resultados se vuelcan en el vector siguiente
	void euler(); // Euler
	void rk4(); //Runge-Kutta de orden 4 
	void configura(const std::vector< std::vector<T> >& a, const std::vector<T>& b); //configuración general
	void configura(); // configuración básica de variables internas 
	
	int actualiza();
public:
	Continuo(unsigned int t, int id, const std::vector< std::vector<T> >& a, const std::vector<T>& b);
				// Constructor de clase. Frecuencia, id, matriz asociada y vector de términos independientes
	Continuo(unsigned int t, int id); // Constructor sin inicialización de vectores


	void propiedades(); // Propiedades de Continuo
	void propiedades_base(); // Propiedades base 
	
};


// Clase controlador --- Clase base de controladores



template <typename T>
class Controlador : public Dinamico<T> { // Controladores: dispositivos dinámicos conectables. 
									  // Ecuaciones en diferencias lineales: implementación preferente para controladores.
									  // Pueden manejar cualquier tipo de número aunque en este trabajo nos centraremos
									  // en escalares en coma flotante (float). 
									  
protected:

	std::promise<unsigned long int> estado_ejecucion;    // Elemento "promesa" para la comunicación asíncrona
										           // Se utiliza para evitar la espera al ejecutar std::thread.join()
										           // el hilo que finalizó todo proceso emite un
												   // estado de ejecución DETENIDO que se obtiene con un objeto
												   // del tipo std::future en el hilo que pretende utilizar
												   // std::thread.join
	std::future<unsigned long int> futuro_estado_ejecucion;  
														// Elemento "futuro". Comunicación asíncrona
													   // Si la "promesa" escribe el estado, el "futuro" lo lee
													   // para dar información a otro hilo, en este caso, del estado
													   // de ejecución
	// Un ejemplo del uso de estas dos variables se halla en el vínculo controlador-watchdog, en el que el controlador
	// puede ser detenido por el watchdog, pero a su vez no puede esperar cuando ejecute join(); necesita una información
	// "a tiempo real" del estado del hilo que cierra. 


	Watchdog<T> *watchdog=nullptr; 
		// Puntero al perro guardián del dispositivo. Se asume la posibilidad
		// de esta clase de supervisión en el caso de los controladores.
	void configura(const std::vector<unsigned int>& v); // Configurador 

				// Nota: Una buena práctica en programación, es asignar los punteros que no se usan inicialmente
				// a un valor de puntero nulo, conocido tradicionalmente en C como NULL. 
				// Sin embargo, en c++11 se ha normalizado la palabra clave nullptr para estos punteros nulos.
	int actualiza(); // Actualizador
	void rutina(); // Rutina del thread
	
public:
	Controlador(unsigned int t, int id, const std::vector<unsigned int>& v, Watchdog<T> *watchdog); // Constructor 1
	Controlador(unsigned int t, int id, Watchdog<T> *watchdog); 									// Constructor 2
	std::future<unsigned long int>& obtiene_estado(); 
						// Obtiene el estado de ejecucion del hilo, reflejado en estado hilo
						 // mediante un objeto std::shard_future de comunicación asíncrona compartida
	std::thread& acaba_forzado(int& error); // Detención forzada
	void propiedades();  // Propiedades
	void propiedades_base();
	
};

// Controlador PID. Uso de la clase Dinamico para crear retardos z^-1 en el dominio discreto. De esta forma se pueden
// implementar toda clase de controladores, basados en ecuaciones en diferencias. 


template <typename T>
class Control_PID : public Controlador<T> {
	
		T p; // Parámetros PID
		T i;
		T d;
	
	T a; // Coeficientes internos
	T b;
	T c;
	
	std::atomic<bool> modificado {false}; // ¿Se modificaron los parámetros?
	
		T p_m; // Parámetros PID modificados
		T i_m;
		T d_m;
	
	float period; // Periodo de muestreo; 
	
	
				// Constantes PID (modificables en ejecución, aunque hay que especificar sus valores iniciales)

	
protected:
	T x = 0; // Señal de entrada al PID
	T y = 0; // Señal de salida al PID ----> El esquema de este PID está en la memoria del trabajo
				// F. de transferencia básica en z: G(z) = p + iT/2(z+1/z-1) + d/T(z-1/z)
				// Para ser utilizable se ha de convertir en ecuaciones en diferencias. Denominador común
				// y descomposición en factores/transformadas inversas simples, con G(z) = Y(z) / X(z) 
	
	void configura(const std::vector<unsigned int>& v); // Configurador 
	T pid();    // método PID. Controlador básico (integración trapezoidal)

	int actualiza(); // Actualizador
	void rutina(); // Rutina del thread	

public:	
	Control_PID(unsigned int t, int id, Watchdog<T> *watchdog, T p, T i, T d);
	void propiedades(); // Propiedades del controlador PID	
	void propiedades_base(); // Propiedades base 
	
};


class Bus; // Necesaria esta declaración de Bus, para ser reconocida por Discretizador (Forward Declaration)

class Discretizador : public Conectable { // Objetos discretizadores. No dinámicos. Conectables. 
										// Estos dispositivos tienen un solo puerto hacia los dispositivos de planta o
										// controladores. No necesitan conocer la estructura de los datos que procesan
	Bus *bus = nullptr; // Bus de comunicaciones 
	
	// Datos revisables en propiedades por el usuario
	std::atomic<int> err{0}; // Error o información proporcionada por el bus 
	std::atomic<int> retorno_bus{0}; // Código de error o condición de lectura/escritura efectuada en Bus
	std::atomic<bool> escribe_a_dispositivo{false}; // Si está el puerto en modo solo lectura (no llegó un paquete nuevo
													// desde el bus) el dispositivo no escribe y sería false
	plantilla_datos_discretizador_sv parametros; // parametros del discretizador (trasmitibles por bus).
		
	// NOTA: La condición de alarma del discretizador se activa si no hay línea de cola de mensajes efectiva
	// al emitir/recibir

protected:

	bool hay_bus = true; // ¿Hay bus asociado?
	
	int esid_conexion = 0; // esid y longitudes de conexión accesible por clases derivadas HW1 y HW2
	unsigned int lal_bus_conexion = 0;
	unsigned int lal_dispositivo_conexion = 0;
	
	BYTE *al_bus = nullptr;		// Puntero a datos al bus, al dispositivo (el discretizador actúa de intermediario)
	BYTE *al_dispositivo = nullptr;
	
	BYTE *al_bus_c = nullptr;		// Punteros a datos al bus y al dispositivo, incluyendo conteo y flag de estado		
	BYTE *al_dispositivo_c = nullptr;		

	int actualiza(); //actualizador y rutina del thread 
	void rutina();	
								
public:
	Discretizador(unsigned int t, int id, unsigned int lcola, Bus *bus); // Constructor del discretizador
										// Los datos de E/S los conocerá del dispositivo mediante el conector
										 // La id ha de coincidir con el discretizador al otro lado del bus.
										 // lcola es la máxima longitud de cola del discretizador (lcola > 1) 
	int asociar(Conector *conector);  // Asociamos a un dispositivo enlazado a un conector, puerto 1, y al bus 
	plantilla_datos_discretizador_sv obtiene_parametros(); // Obtener parametros del discretizador															
													  // Este método es necesariamente usado por el bus 
	~Discretizador(); // Destructor. Necesario para liberar la memoria de los buffers 
	void propiedades();  //Propiedades en pantalla
	void propiedades_base();
	
};

// Clases derivadas del discretizador. Estas clases intercambian entrada/salida de forma física
// Haciendo uso de los puertos. Las clases que se deriven para uso específico necesitan
// 1) Conocer la estructura de datos E/S que captarán del controlador
// 2) Inicializarse convenienemente (requisitos HW de iniciación, puertos, etc.
// 3) Utilizar su rutina hardware() para realizar la entrada/salida
// 4) Finalizar con el destructor de tal forma que se puedan cerrar los puertos convenientemente

class Discretizador_HW1 : public Discretizador // Discretizador HW1 o discretizador sin conector
{											// Utilizado para intercambio E/S con dispositivo físicos
											// Mediante un proceso de planta 	
		
protected:										
	
	int actualiza(); //actualizador y rutina del thread 
	void rutina(); // Rutina de hilo 
	virtual void hardware(); // Rutina de E/S con el hardware
	
public:	 
		// Constructor. Requiere parámetros de los paquetes de datos: esid y longitudes; también el bus
	Discretizador_HW1(unsigned int t, int id, int esid, unsigned int laplanta, unsigned int lacontrolador, 
								unsigned int lcola, Bus *bus);
	~Discretizador_HW1(); // Destructor
	int asociar();  // Asociamos al bus. No hay conector. 
};


class Discretizador_HW2 : public Discretizador { // Discretizador HW2 o discretizador sin bus
											// Utilizado para intercambio E/S con dispositivos físicos
											// Mediante un proceso de planta
	
protected:											

	int actualiza(); //actualizador y rutina del thread 
	void rutina();	// Rutina de hilo
	virtual void hardware(); // Rutina de E/S con el hardware
											  
public:
		// Constructor. Requiere parámetros de los paquetes de datos: esid y longitudes. Sin bus
	Discretizador_HW2(unsigned int t, int id, int esid, unsigned int laplanta, unsigned int lacontrolador);	
											
	int asociar(Conector *conector);  // Asociamos a un dispositivo enlazado a un conector, puerto 1. No hay bus									
};

class Bus : public Dispositivo {  // Comunicación por cola de mensajes entre planta y controlador
								  // Se utiliza estándar POSIX de mensajería. Por cada par conectado hay un canal
								  // de entrada y otro de salida. Estos buses, además, almacenan en forma de cola
								  // todos los mensajes 
								  
	key_t clave_cola; // Clave de la cola, se utilizará en envío y recepción de mensajes. Depende de la id del bus
					  // Referencia de archivo: "." (directorio actual). planta y controlador deberán estar en el 
					  // mismo directorio para que se establezca la cola.	
					  
	struct plantilla_operacion_bus { // Operación en el bus del discretizador (flags)
			bool enviar_cabecera; // Ha de enviar cabecera
			bool recibir_cabecera; // Ha de recibir cabecera
			bool recibida_cabecera; // La cabecera fue recibida por el propio bus
			bool recibida_cabecera_otro; // La cabecera fue recibida por el otro bus
			bool incompatible; // Se detectó incompatibilidad entre discretizadores (grave)
			bool difieren_colas; // difieren colas?
			bool difieren_fmuestreo; // difiere frecuencia de muestreo?
			unsigned int conteo; // conteo local
			unsigned int conteo_otro; // conteo de la otra cola
			BYTE *al_bus_c; // punteros de datos
			BYTE *al_dispositivo_c;
	};
	
	enum tipo_proceso tipo; // Tipo de bus que emplearemos (lado del controlador o lado de la planta).
	int id_cola; // Identificación de cola
	bool conexion = false; // ¿hay posibilidad de conexión con el otro bus? 
	
	std::mutex exclusion_bus; // mutex para las operaciones del bus
	
	std::vector<plantilla_datos_discretizador_sv> parametros; // Parametros del discretizador local
	std::vector<plantilla_datos_discretizador_sv> parametros_otro; // Parametros que emite el otro discretizador
	
	std::vector<plantilla_operacion_bus> operacion; // Operación del discretizador
	
	unsigned int busca(int id); // buscar según id
	
	
public:
	Bus(int id, tipo_proceso tipo); // Constructor. Necesita la id de bus y el tipo de proceso (controlador o planta). 
					// Es un elemento pasivo sin hilo propio. 
					// Son los discretizadores asociados los que leen y escriben en la cola de mensajes del bus
					// el ID ha de coincidir con el otro bus para conectar
	
														 
	int registrar(Discretizador *discretizador, BYTE *al_bus_c, BYTE *al_dispositivo_c); // Registrar discretizador
																		// y también zonas de datos completas
																		// que incluyen conteo de cola
	
	int lecturaescritura(Discretizador *discretizador, bool solo_lectura, int *err_es);  
												// Lectura/escritura en el bus. El bus devuelve información según	
												// su naturaleza, tanto en err_es (error efectivo E/S). 
												// y también por retorno (generalmente datos efectivos leidos/escritos)
												
												// Con el flag solo_lectura, no enviará al otro bus. Como en el 
												// caso de los conectores, con ello se garantiza que se enviarán
												// datos válidos. 
															
	~Bus(); // Destructor. Se ha de cerrar el sistema de colas con msgctl(Bus::id_cola, IPC_RMID, 0)

	void propiedades(); // Propiedades del bus. 
	void propiedades_base(); // propiedades base
	
};







// Clase Interactivo - Interacción con el usuario (mostrar propiedades, encender/apagar dispositivos/crear gráficas,etc.


class Interactivo { // Clase NO derivada de Dispositivo. Es el menú interactuador de la planta-controlador
					// Permite al usuario tener un control del sistema, grabar  e interactuar cambiando variables
					
		enum plantilla_primer_nivel {INICIAR, DETENER, PAUSAR, REANUDAR}; // Control de ejecución en el menú		
		
															// Opciones primer nivel.
															// No necesariamente se muestran todas estas opciones y
															// dependerán del contexto. Por ejemplo, si todos los hilos
															// se están ejecutando, es innecesario mostrar la opción
															// "Arrancar todos". Esta comprobación lo hace la rutina
															// conformar_menu_primer_nivel();
		std::vector<Dispositivo *> dispositivos; // Lista de dispositivos
		std::vector<std::string> nombres; // Nombres asignados a los dispositivos, por el usuario
//		std::vector<std::vector<plantilla_caracteristicas>> caracteristicas; // Vector de vectores de características
																	    // Para el menú 'propiedades'
		std::vector<bool> ejecutable; // Indicar si es ejecutable para el menu correspondiente de
								      // donde se puede elegir inicio-pausa-retomar-parada
		std::vector<bool> ejecucion; // ¿Se está ejecutando?
		std::vector<bool> pausa; //¿ Está pausado?
		std::vector<bool> alarma; //¿ hay alarma?
		
	
		
		std::vector<plantilla_primer_nivel> opciones_1; // Vector primer nivel de opciones (general), que es
															   // variable según el contexto.
															   
//		std::vector<enum Clase> opciones_2; // Vector segundo nivel de opciones (dispositivo), que es
															   // variable según el contexto.
		
		
		void conformar_menu_primer_nivel(); // Actúa sobre el vector de opciones del primer nivel
											// en función de los estados de todos los dispositivos registrados
		
//		void conformar_menu_segundo_nivel(unsigned int dispositivo); 
											// Actúa sobre el vector de opciones del segundo nivel
											// en función de las propiedades y características de cada dispositivo
		
		void arrancar_todos(); // Acciones generales de dispositivos, en el menú principal
		void detener_todos();
		void pausar_todos();
		void retomar_todos();
		
		void refresca_estados();//Se refrescan los estados de todos los dispositivos 
								// en los vectores(ejecucion, pausa, etc.)
								
		unsigned int menu_general();  // Menú general (Nivel 1)		
		void menu_dispositivo(unsigned int dispositivo); // Menú de dispositivo (Nivel 2)
public:
//		Interactivo(); // Constructor
		int registro(Dispositivo *dispositivo, std::string nombre); //registro de dispositivo en el menú interactivo 
		int ejecutar(bool borrar); // Ejecutar menú. Flag 'borrar' a true permite destruir objetos

		
		
		
};

#endif // BASE_H
