// CONTROLADOR.CPP 
//
// Implementación de los controladores, derivado de conectables y sistemas dinámicos 
// Incluye controladores PID y se pueden añadir otros tipos de controladores
//
// Estos dispositivos tienen como base (clase controlador) una conexión con el dispositivo watchdog
//
//

#include "base.h" // declaración de la clase base

#include <iostream> // Para salida por pantalla

// Clase controlador --- Clase base de controladores 

template <typename T>
Controlador<T>::Controlador(unsigned int t, int id, const std::vector<unsigned int>& v, Watchdog<T> *watchdog) 
																: Dinamico<T>(t, id, v) { // Constructor 1

	Controlador<T>::watchdog = watchdog; // Inscribimos watchdog
	
}

template <typename T>
Controlador<T>::Controlador(unsigned int t, int id, Watchdog<T> *watchdog) : Dinamico<T>(t, id) { // Constructor 2
																							// Colas "a posteriori"
	

	Controlador<T>::watchdog = watchdog; // Inscribimos watchdog
	
}


template <typename T>
void Controlador<T>::configura(const std::vector<unsigned int>& v) { // Configurador que crea las colas "a posteriori"
	
	Dinamico<T>::configura(v);

}

template <typename T>
std::future<unsigned long int>& Controlador<T>::obtiene_estado() { // obtener estado de ejecución por future
	
	return Controlador<T>::futuro_estado_ejecucion;

}


template <typename T>
int Controlador<T>::actualiza() { 
							// Actualización del controlador 
	
	int resultado=0;	
	
	if(Controlador<T>::watchdog != nullptr)	
		Controlador<T>::watchdog -> sigo_operativo(this); 
							// Se notifica al watchdog que el controlador sigue operativo	
	
	
	resultado += Dinamico<T>::actualiza(); // Actualizar clase base
	return resultado;

}

template <typename T>
void Controlador<T>::rutina() { // rutina thread del watchdog. Se limita a repetir un ciclo hasta el cese del hilo
	
	Controlador<T>::estado_ejecucion = std::promise<long unsigned int>(); 
									// Se borra promesa anterior igualando a constructor vacío 
	
	Controlador<T>::futuro_estado_ejecucion = Controlador<T>::estado_ejecucion.get_future(); // Adquirimos futuro
	
	Controlador<T>::estado_ejecucion.set_value_at_thread_exit(CODIGO_FIN_HILO);
	
	
	while (Dispositivo::obtiene_ejecucion()) 		
		Controlador<T>::actualiza();	
	
}

template <typename T>
std::thread& Controlador<T>::acaba_forzado(int& error) { // Detención forzada del hilo (para watchdog)		
													     // Desactiva la ejecución definitivamente por entenderse
														 // que el controlador no funciona adecuadamente
	
	std::thread& hilo = Dispositivo::acaba(error);  // Acabamos de forma habitual
		
	Dispositivo::desactiva_ejecucion(); // Intentamos desactivar (si no estuviera acabado, no se desactivará)
	
	return hilo;
}

template <typename T>
void Controlador<T>::propiedades() { // propiedades

	
}

template <typename T>
void Controlador<T>::propiedades_base() { // propiedades base (dinamico)

	Controlador<T>::propiedades();

	Dinamico<T>::propiedades_base();
	
	return;
}


// Control PID ---------------------------------------- CONTROLADOR DE AMPLIO USO. La integración es trapezoidal
		

template <typename T> 
Control_PID<T>::Control_PID(unsigned int t, int id, Watchdog<T> *watchdog,  T p, T i, T d) 
																: Controlador<T>(t,id,watchdog){
											// Constructor del controlador PID
	float period = Dispositivo::obtiene_period(); // Hallamos período de muestreo T

	// Crear vector de colas para el sistema dinámico y reconfigurar la clase correspondiente
	std::vector<unsigned int> v{2,1};// Cola tamaño 2 + Cola tamaño 1.
									 //Suficiente para satisfacer las ecuaciones en diferencias 
						//					u[k] = u[k-1]+a*e[k]+b*e[k-1]+c*e[k-2]
						//  		        y = u[k] ; x = e[k] (Adjunto en la memoria del trabajo)
	
	
	Control_PID<T>::p = p; // Parámetros PID 
	Control_PID<T>::i = i;
	Control_PID<T>::d = d;
	
	Control_PID<T>::p_m = p; // Parámetros PID (modificables)
	Control_PID<T>::i_m = i;
	Control_PID<T>::d_m = d;
	
	
	Control_PID<T>::configura(v); // Configuramos 
	
	// Aviso: Es labor de la clase derivada modificar
	// las condiciones iniciales si fuese conveniente, se modificarían en clases derivadas. Por defecto son cero. 
	
	Dinamico<T>::siguiente[0] = 0; //  y = e[k] Estado inicial a cero. 
	Dinamico<T>::siguiente[1] = 0; //  x = u[k] Estado inicial a cero. 
	
	Control_PID<T>::period = period; // Queda permanente el período de muestreo
	
	// Calculamos coeficientes 
	
	Control_PID<T>::a = Control_PID<T>::p + Control_PID<T>::i*(T)(Control_PID<T>::period/2) +
												Control_PID<T>::d/(T)(Control_PID<T>::period);
	Control_PID<T>::b = -Control_PID<T>::p + Control_PID<T>::i*(T)(Control_PID<T>::period/2) -
												(2*Control_PID<T>::d/(T)(Control_PID<T>::period));
	Control_PID<T>::c = Control_PID<T>::d / (T)(Control_PID<T>::period);
	
	
	return;
	
}

template <typename T>
void Control_PID<T>::configura(const std::vector<unsigned int>& v) { // Configurador (llamada a Controlador)	
	
	Controlador<T>::configura(v); // Se crean las colas	
	
	return;
	
}

template <typename T> // Controlador PID estándar
T Control_PID<T>::pid() {
	
	T valor;
	
	if(Control_PID<T>::modificado) { // ¿Se modificaron los parámetros? Se recalcula
	
		Control_PID<T>::p = Control_PID<T>::p_m; // Modificamos los parámetros pid según lo seleccionado en propiedades()
		Control_PID<T>::i = Control_PID<T>::i_m;
		Control_PID<T>::d = Control_PID<T>::d_m;		
	
		Control_PID<T>::a = Control_PID<T>::p + Control_PID<T>::i*(T)(Control_PID<T>::period/2) +
												Control_PID<T>::d/(T)(Control_PID<T>::period);
		Control_PID<T>::b = -Control_PID<T>::p + Control_PID<T>::i*(T)(Control_PID<T>::period/2) -
												(2*Control_PID<T>::d/(T)(Control_PID<T>::period));
		Control_PID<T>::c = Control_PID<T>::d / (T)(Control_PID<T>::period);
		Control_PID<T>::modificado = false; // Ya se modificó. 
	}
	
	// Aplicamos ecuación en diferencias u[k] = u[k-1]+a*e[k]+b*e[k-1]+c*e[k-2]
	Control_PID<T>::y = Dinamico<T>::estado[1][0]+Control_PID<T>::a*Control_PID<T>::x+
		Control_PID<T>::b*Dinamico<T>::estado[0][0]+Control_PID<T>::c*Dinamico<T>::estado[0][1];
		
//	std::cout << Control_PID<T>::x << "\n";
		
	Dinamico<T>::siguiente[0] = Control_PID<T>::x; // Actualizamos estados
	Dinamico<T>::siguiente[1] = Control_PID<T>::y;
	
	valor = Control_PID<T>::y;	// Se retorna con el nuevo estado y[k] 
				
	return valor;
	
}

template <typename T>
int Control_PID<T>::actualiza() { // Actualización
	
	int resultado=0;	
	
	resultado += Controlador<T>::actualiza(); // Actualizar clase base
	return resultado;
	
}

template <typename T>
void Control_PID<T>::rutina() { // rutina thread del watchdog. Se limita a repetir un ciclo hasta el cese del hilo
	
	Controlador<T>::estado_ejecucion = std::promise<long unsigned int>(); 
									// Se borra promesa anterior igualando a constructor vacío 
	
	Controlador<T>::futuro_estado_ejecucion = Controlador<T>::estado_ejecucion.get_future(); // Adquirimos futuro
	// ESTA LINEA SE PUEDE COMENTAR PARA TESTAR EL MECANISMO DE EXCEPCIONES
	
	Controlador<T>::estado_ejecucion.set_value_at_thread_exit(CODIGO_FIN_HILO);
	
	while (Dispositivo::obtiene_ejecucion()) 		
		Control_PID<T>::actualiza();	
	
}

template <typename T>
void Control_PID<T>::propiedades() { // propiedades del controlador PID

	unsigned int opcion; 
	float p, i, d;
	bool editable = false; //¿ Es editable?

	do {
		std::cout << "\n\n-------- PROPIEDADES DEL CONTROLADOR PID ----------------------------\n\n";
	
		std::cout << "\t\tP:\t" << Control_PID<T>::p_m;
		std::cout << "\n\t\tI:\t" << Control_PID<T>::i_m;
		std::cout << "\n\t\tD:\t" << Control_PID<T>::d_m;
		
		if(!Control_PID<T>::modificado && Dispositivo::obtiene_ejecucion()) { // Mostrar opciones si son modificables 
			std::cout << "\n\n\t1) Modificar valor P";
			std::cout << "\n\t2) Modificar valor I";
			std::cout << "\n\t3) Modificar valor D";
			editable = true;
		} else {
			std::cout << "\n\nLos valores no son modificables en este momento, hasta que los procese el dispositivo."
					  << "\nSi el dispositivo está parado/pausado hay que iniciarlo para hacer efectivos los cambios" 
					  << "\nPulse cualquier valor que no sea 0 para actualizar esta pantalla o 0 para salir.";
			editable = false;
		}
		std::cout << "\n\n\t0) Salir";
		
		std::cout << "\n\n---------------------------------------------------------------------\n";
		
		std::cout << "\nSeleccione opción -> "; 
		std::cin >> opcion; //espera orden de teclado			 
		
		if(editable) { // ¿Es modificable?
			switch(opcion) {
				case 1:
					std::cout << "\n\nIntroduzca nuevo valor P -> ";
					std::cin >> p; //espera orden de teclado
					Control_PID<T>::p_m = p; 				
					Control_PID<T>::modificado = true; // ¿Se modificaron valores? si
					break;
				case 2:
					std::cout << "\n\nIntroduzca nuevo valor I -> ";
					std::cin >> i; //espera orden de teclado
					Control_PID<T>::i_m = i; 
					Control_PID<T>::modificado = true; // ¿Se modificaron valores? si
					break;
				case 3:
					std::cout << "\n\nIntroduzca nuevo valor D -> ";
					std::cin >> d; //espera orden de teclado
					Control_PID<T>::d_m = d; 
					Control_PID<T>::modificado = true; // ¿Se modificaron valores? si
					break;
				default:
					break;		
			};	
		} 

	} while (opcion != 0);	
		
	return;
}

template <typename T>
void Control_PID<T>::propiedades_base() { // propiedades base (Se indica base superior que es Controlador)

	Controlador<T>::propiedades_base();	
	
}


template class Controlador<float>; // Plantilla que se usará. Se ha de declarar aquí, de lo contrario hay error de linker
template class Control_PID<float>; // Plantilla que se usará. Se ha de declarar aquí, de lo contrario hay error de linker