// MOTORCC_ARDUINO.CPP
//
// Implementación de un controlador motorCC en Arduino.

#include "motorcc_arduino.h"

// Clase Hardware. Entrada/salida con el motor, utilizando convertidores DAC y ADC integrados en el microcontrolador


// Constructor Hardware. Inicializa E/S según los canales proporcionados en los argumentos

Hardware::Hardware(unsigned int canalE, unsigned int canalS, unsigned int canal_consigna, unsigned int canal_alarma) {
	
// Configurar Watchdog hardware de Arduino
	
// Configurar pines de entrada según canal
		int entrada = 0, consigna = 0; // Entradas analógicas (A0, A1, A2, A3, A4, A5)
		
		switch(canalE) // Se asignan los canales analógicos según casos
		{
			case 0:
				entrada = A0;
				break;				
			case 1:
				entrada = A1;
				break;	
			case 2:
				entrada = A2;
				break;	
			case 3:
				entrada = A3;
				break;	
			case 4:
				entrada = A4;
				break;	
			case 5:
				entrada = A5;
				break;					
		};
		
			switch(canal_consigna)
		{
			case 0:
				consigna = A0;	
				break;	
			case 1:
				consigna = A1;
				break;	
			case 2:
				consigna = A2;
				break;	
			case 3:
				consigna = A3;
				break;	
			case 4:
				consigna = A4;
				break;	
			case 5:
				consigna = A5;
				break;	
		};
		
		if (consigna == entrada) return; // No se admiten canales iguales de entrada analogica 
		
		Hardware::canalE = entrada; // Se guardan canales analogicos (canal de salida en PWM)
		Hardware::canal_consigna = consigna;
		
		Hardware::canalS = canalS;
		
		pinMode(canal_alarma, INPUT); // Se configura canal de alarma como entrada
		
		return;
	
}

// Conocer si el controlador está en estado de alarma

bool Hardware::estado_alarma() {
	
	return Hardware::alarma;	
}

int Hardware::comunicacion_hardware() { // Comunicación de hardware (E/S) y acondicionamiento de señales

	if(Hardware::canalE == 9999 || Hardware::canal_consigna == 9999) return -1; 
								// No se configuró bien, retorno sin hacer nada
	// Entradas y salida analógica (tacómetro, consigna y voltaje de salida) en formato entero y coma flotante							
	int tac_i, con_i, vol_i; 
	float tac_f, con_f, vol_f;
	
	bool alarma = false; // Alarma		
	
	// Lectura de tacómetro y consigna (entradas)
	tac_i = analogRead(Hardware::canalE); // Se lee puerto analógico de entrada
	con_i = analogRead(Hardware::canal_consigna); // Se lee puerto analógico de entrada
	
	// Conversión, de entero a float del valor del tacómetro. 
	// Sabemos sus límites en coma flotante con TAC_MAX y TAC_MIN, por tanto, convertimos a decimal su valor		
	tac_f = MIN_TAC + (MAX_TAC - MIN_TAC) * ((float)tac_i / (float)ESCALA_ADC);
	
	// Conversión, de entero a float del valor de la consigna. 
	// Sabemos sus límites en coma flotante con CON_MAX y CON_MIN, por tanto, convertimos a decimal su valor		
	con_f = MIN_CON + (MAX_CON - MIN_CON) * ((float)con_i / (float)ESCALA_ADC);
	
	// Se actualizan valores para ser leidos por las subclases
	Hardware::entrada = tac_f; 
	Hardware::consigna = con_f; 
	
	// Conversión de salida (voltaje)
	vol_f = Hardware::salida;
	
	vol_f = vol_f > MAX_VOL ? MAX_VOL : vol_f; // Comprobamos límites que no se pueden sobrepasar
	vol_f = vol_f < MIN_VOL ? MIN_VOL : vol_f;
	
	// Conversión, de float a entero del voltaje. 
	// Sabemos sus límites en coma flotante con VOL_MAX y VOL_MIN, por tanto obtenemos su valor
	
	// convertimos en tanto a uno con respecto a los límites
	
	vol_f = (vol_f - MIN_VOL) / (MAX_VOL - MIN_VOL); 
	
	// Convertimos a entero
	
	vol_i = (int)((float)ESCALA_DAC * vol_f);
	
	// Escribimos en el puerto de salida
	
	analogWrite(Hardware::canalS, vol_i);

	// Leemos la alarma
	
	alarma = digitalRead(canal_alarma) == HIGH ? true : false; // Se lee alarma estado HIGH = true
	
	Hardware::alarma = alarma; // Se actualiza valor de alarma en la memoria RAM
	
	return 0;

}

int Hardware::actualiza() { // Actualización por cada iteración de Hardware
	
	return Hardware::comunicacion_hardware(); // Llamada a E/S y actualización de registros
	
}

// Clase DINAMICO. Retención de estados anteriores y desplazamiento de estos en cada iteración (al actualizar)

// Constructor de la clase, argumentos a la clase base Hardware
Dinamico::Dinamico(unsigned int canalE, unsigned int canalS, unsigned int canal_consigna, unsigned int canal_alarma,
		unsigned int longitud1, unsigned int longitud2, unsigned int longitud3, 
			unsigned int longitud4) : Hardware(canalE, canalS, canal_consigna, canal_alarma) {
				
	unsigned int i, j; // Indices
				
	unsigned int longitud1_ = longitud1; // Asignamos las colas
	unsigned int longitud2_ = longitud1;
	unsigned int longitud3_ = longitud1;
	unsigned int longitud4_ = longitud1;
				
	if (longitud1_ > MAX_COLA) longitud1_ = MAX_COLA; // Se limitan las colas
	if (longitud2_ > MAX_COLA) longitud2_ = MAX_COLA;
	if (longitud3_ > MAX_COLA) longitud3_ = MAX_COLA;
	if (longitud4_ > MAX_COLA) longitud4_ = MAX_COLA;
				
	Dinamico::longitudes[0] = longitud1_; // Cargamos longitudes de cada cola en el vector de colas
	Dinamico::longitudes[1] = longitud2_;
	Dinamico::longitudes[2] = longitud3_;
	Dinamico::longitudes[3] = longitud4_;
	
	for(i = 0 ; i < 4 ; i++) { // Rellenamos colas con ceros
		for(j = 0 ; j < longitudes[i] ; j++) {
			switch(i) {
				case 0:
					Dinamico::cola1[j] = 0.;
					break;	
				case 1:
					Dinamico::cola2[j] = 0.;
					break;	
				case 2:
					Dinamico::cola2[j] = 0.;
					break;	
				case 3:
					Dinamico::cola2[j] = 0.;
					break;					
			};
		}
	}
				
	return;			
				
}

int Dinamico::actualiza() { // Actualización por cada iteración de Dinamico

	unsigned int i, j; // Indices
	

	for(i = 0 ; i < 4 ; i++) { // Desplazamiento de cada cola
		for(j = 0 ; j < longitudes[i]-1 ; j++) {
			switch(i) {
				case 0:
					Dinamico::cola1[j+1] = Dinamico::cola1[j];
					break;	
				case 1:
					Dinamico::cola2[j+1] = Dinamico::cola2[j];
					break;	
				case 2:
					Dinamico::cola3[j+1] = Dinamico::cola3[j];
					break;	
				case 3:
					Dinamico::cola4[j+1] = Dinamico::cola4[j];
					break;					
			};
		}
	}
	
	for(i = 0 ; i < 4 ; i++) { // Sustitución del nuevo valor de estado		
		switch(i) {
			case 0:
				if(longitudes[i] > 0)
					Dinamico::cola1[0] = estado_actual[0];
				break;	
			case 1:
				if(longitudes[i] > 0)
					Dinamico::cola2[0] = estado_actual[1];
				break;	
			case 2:
				if(longitudes[i] > 0)
					Dinamico::cola3[0] = estado_actual[2];
				break;	
			case 3:
				if(longitudes[i] > 0)
					Dinamico::cola4[0] = estado_actual[3];
				break;					
		};
	}	
	
	return Hardware::actualiza(); // Llamada a clase base
	
}

// Clase Controlador PID. Se usarán dos colas: una de longitud dos y otra de longitud uno (ver capítulo 4 de la memoria)

Controlador_PID::Controlador_PID(unsigned int canalE, unsigned int canalS, // Constructor
	unsigned int canal_consigna, unsigned int canal_alarma,	float p, float i, float d, float t) : Dinamico(canalE, 
	canalS, canal_consigna, canal_alarma, 2, 1, 0, 0)	{
		
	Controlador_PID::p = p; // Guardamos coeficientes PID
	Controlador_PID::i = i;
	Controlador_PID::d = d;
		
	Controlador_PID::a = p + (i*t/2) + (d/t); // Calculamos coeficientes de las ecuaciones en diferencias
	Controlador_PID::b = -p + (i*t/2) - (2*d/t);
	Controlador_PID::c = d/t;		
		
	return;	
}

int Controlador_PID::actualiza() { // Actualización en cara iteración. Se resuelven ecuaciones en diferencias

	// Calculamos la diferencia en el lazo de control (error e, al restar consigna y tacómetro)
	
	float e = Hardware::consigna - Hardware::entrada;
	
	// Aplicamos ecuación en diferencias u[k] = u[k-1]+a*e[k]+b*e[k-1]+c*e[k-2]
	
	float u = Controlador_PID::cola2[0] + Controlador_PID::a * e +
		Controlador_PID::b * Controlador_PID::cola1[0] + Controlador_PID::c * Controlador_PID::cola1[1];
		
	Dinamico::estado_actual[0] = e; // Actualizamos estados
	Dinamico::estado_actual[1] = u;

	Hardware::salida = u; // La salida del controlador PID es el voltaje
	
	return Dinamico::actualiza(); // Llamada a clase base
}

