// WATCHDOG.CPP
//
// Clase de dispositivos de supervisión o watchdog (perro guardian). Están únicamente asociados a controladores
// Los controladores están obligados a hacer una llamada de control al watchdog, cada cierto tiempo.
// Si el tiempo entre llamadas es superior a lo establecido en el watchdog (un múltiplo de ciclos de muestreo del
// watchdog) entonces éste activa su señal de alarma, indicando al usuario que el controlador no funciona correctamente

#include "base.h"
#include <iostream>

// Implementaciones de clase de watchdog (vigilancia de controladores) 
// Cada watchdog puede albergar varios controladores. Cuando uno de los controladores falla (al no hacer la llamada
// regular pertinente) se pone el watchog en modo alarma. En propiedades se puede comprobar qué dispositivo ha 
// fallado 

template <typename T>
Watchdog<T>::Watchdog(unsigned int t, int id) : Dispositivo(t,id) { 
		
		// Constructor del watchdog

		// Se recomienda que el Watchdog tenga la misma frecuencia de muestreo que el controlador o controladores
		// El argumento de conteo está basado en esta frecuencia y hay que tenerlo en cuenta cuando se registra
		// un controlador de frecuencia menor																	

}

template <typename T>
int Watchdog<T>::registrar(Controlador<T> *controlador, unsigned int multiplo, bool detencion) { 
		// Registro de controlador, pasa puntero del controlador, multiplo de conteo para alarma y bit de detención	
		// true = detiene el controlador; false = no detiene el controlador
	
	int err = 0; // Código de retorno
	plantilla_controlador registro_controlador;	
	
	if(controlador == nullptr) return -1; // No se pueden pasar punteros nulos
	
	registro_controlador.controlador = controlador;
	
	registro_controlador.id = controlador -> obtiene_id();
	registro_controlador.detencion = detencion; 
	registro_controlador.estado_detencion = CONTROLADOR_NORMAL;
	registro_controlador.fmuestreo = controlador -> obtiene_muestreo();	
//	registro_controlador.estado_ejecucion = controlador -> obtiene_estado(); // Obtiene future del estado
	
	if(multiplo > 0)
		registro_controlador.multiplo = multiplo; 
							//múltiplo o número de períodos de muestreo que no se pueden sobrepasar
	else {
		registro_controlador.multiplo = 1; 
						// el valor de múltiplo no puede ser cero! (el watchdog se dispararía nada más comenzar)
		err = -2;
	}
	
	registro_controlador.conteo = 0;
	registro_controlador.alarma = false;
	
	Watchdog<T>::controladores.push_back(registro_controlador);
	
	return err; // Código de retorno
}

template <typename T>
int Watchdog<T>::actualiza() { 
							// Actualización del watchdog. Incrementa conteo y si éste supera a multiplo: alarma
	
	int resultado=0;
	register unsigned int i, indice; //indices
	
	for(i = 0 ; i < Watchdog<T>::controladores.size(); i++) { // Recorremos todo el vector de controladores
		if(!Watchdog::controladores[i].alarma) { // Descartamos controladores con alarma activada																			 
			Watchdog<T>::controladores[i].conteo++; // incrementamos conteo
			
			if(!Watchdog<T>::controladores[i].alarma) { // Recorremos controladores sin alarma  
				if (Watchdog<T>::controladores[i].conteo > Watchdog::controladores[i].multiplo) { 
																				// Si supera múltiplo
					Dispositivo::activa_alarma(); // Activamos alarma del dispositivo watchdog
					Watchdog<T>::controladores[i].alarma = true; // Y la alarma en la entrada correspondiente
					if(Watchdog<T>::controladores[i].detencion) { // ¿Se ha pedido detención del hilo?
						
						Watchdog<T>::hilos_pendientes.push_back(
							{std::move(Watchdog<T>::controladores[i].controlador -> 
							  acaba_forzado(Watchdog<T>::controladores[i].err)),
							  std::move(Watchdog<T>::controladores[i].controlador -> obtiene_estado()),i}); 
										// se detiene, hace uso del vector y se hace uso de err/
							 
						if(Watchdog<T>::controladores[i].err == 0) // Comprobación del código de retorno
							Watchdog<T>::controladores[i].estado_detencion = CONTROLADOR_DETENIENDOSE;
						else { 													// Estaba ejecutándose
							Watchdog<T>::controladores[i].estado_detencion = CONTROLADOR_DETENIDO;
																				   // Estaba parado. Se borra entrada
							Watchdog<T>::hilos_pendientes.pop_back();
						}
					}
						// Nota: Una vez que el controlador está con alarma, aunque se recupere, seguirá con el
						// status de alarma. Si al detenerse, se reinicia, no cambiará su condición de alarma
				}				
			
			}			
		
		}		
		
	}
	
	for (i=0 ; i < Watchdog<T>::hilos_pendientes.size() ; i++) { // Ronda de "uniones" thread.join()
																// De hilos pendientes. Si no se cerraran
																// los destructores provocarían excepción
		indice = Watchdog<T>::hilos_pendientes[i].indice;
		if(Watchdog<T>::controladores[indice].estado_detencion == CONTROLADOR_DETENIENDOSE) {
			int retorno = Watchdog::futuro_preparado(Watchdog<T>::hilos_pendientes[i].futuro);
			if(retorno == 1)  { // ¿Futuro preparado? Si hubo excepción, se hace join igualmente
				if(Watchdog<T>::hilos_pendientes[i].futuro.get() == CODIGO_FIN_HILO) {
																							//¿finalizó?
				Watchdog<T>::hilos_pendientes[i].hilo.join(); // Juntamos el hilo (no debería haber demora)
				Watchdog<T>::controladores[indice].estado_detencion = CONTROLADOR_DETENIDO;					
				}
				
			}
			if(retorno == -1) { // Hubo excepción. Se hace join() igualmente, aunque haya que esperar. 
				Watchdog<T>::hilos_pendientes[i].hilo.join(); // Juntamos el hilo 
				Watchdog<T>::controladores[indice].estado_detencion = CONTROLADOR_DETENIDO;				
			}
		
		}
	}
		
			
	
	resultado += Dispositivo::actualiza(); // Actualizar clase base
	return resultado;

}

template <typename T>
void Watchdog<T>::rutina() { // rutina thread del watchdog. Se limita a repetir un ciclo hasta el cese del hilo

	unsigned int i, indice; // Indices para finalizar hilos pendientes
	
	while (Dispositivo::obtiene_ejecucion()) 		
		Watchdog<T>::actualiza();
	
	// Se ha dado orden de detención del watchdog. Si quedase algún controlador por cerrar ha de hacerlo ahora

	for (i=0 ; i < Watchdog<T>::hilos_pendientes.size() ; i++) { // Ronda de "uniones" thread.join()
																// De hilos pendientes. Si no se cerraran
																// los destructores provocarían excepción
																// al destruirse el watchdog saliendo del proceso
		
		indice = Watchdog<T>::hilos_pendientes[i].indice;		
		if(Watchdog<T>::controladores[indice].estado_detencion == CONTROLADOR_DETENIENDOSE) {
		
			Watchdog<T>::hilos_pendientes[i].hilo.join(); // Juntamos el hilo 
			Watchdog<T>::controladores[indice].estado_detencion = CONTROLADOR_DETENIDO;		
		}			
	} 
	
	Watchdog<T>::hilos_pendientes.clear();
} // Fin de rutina watchdog	

template <typename T>
int Watchdog<T>::futuro_preparado(std::future<long unsigned int> const& futuro) { // Verifica si futuro está preparado
					 // Resultado = 1 ready; resultado=0 not ready; resultado=-1 Algo falló
   int resultado;
   
   try { // Zona de intercepción de excepciones
	   resultado = futuro.wait_for(std::chrono::seconds(0)) == std::future_status::ready ? 1 : 0;
   }
   
   catch (std::future_error& e){ // Sería una excepción de tipo future
	
	   std::cerr << "\n Se ha interceptado una excepción con std::future. Mensaje: " << e.what() << "\n";
	   resultado = -1; // Si no se pudo, se informa 
   }
   
   return (resultado); // si está preparado el std::future, devuelve true 
}

template <typename T>
void Watchdog<T>::sigo_operativo(Controlador<T> *controlador) { 
								// Llamada de comprobación por parte del controlador. El watchdog "resetea" el conteo
	unsigned int i;
	Watchdog<T>::exclusion_watchdog.lock(); // Zona de concurrencia 
	
	
	for(i = 0 ; i < Watchdog<T>::controladores.size(); i++) { // Recorremos vector de controladores
		if(Watchdog<T>::controladores[i].controlador == controlador && !Watchdog<T>::controladores[i].alarma)  
			Watchdog<T>::controladores[i].conteo = 0; // Conteo a cero en la entrada correspondiente		
	}		
	
	Watchdog<T>::exclusion_watchdog.unlock(); // Fin de zona de concurrencia
	
	return;
}

template <typename T>
void Watchdog<T>::propiedades() { // Propiedades del dispositivo
		
	unsigned int i; // indice
	std::vector<plantilla_controlador> controladores; // Copia del vector de controladores
	
		
	Watchdog<T>::exclusion_watchdog.lock(); // Zona de concurrencia 
		
	controladores = Watchdog<T>::controladores; // Copiamos 
		
	Watchdog<T>::exclusion_watchdog.unlock(); // Fin de zona de concurrencia
		
		
	std::cout << "\n\n-------- PROPIEDADES DEL WATCHDOG -----------------------------------\n\n";
	std::cout << "Controlador\tid\tDetener\tfmuestreo\tconteo\tmax.conteo\talarma\terr\tstatus\n\n";
	for(i = 0 ; i < controladores.size(); i++) {
		std::string alarma = controladores[i].alarma ? "SI" : "NO"; // bool a string
		std::string detencion = controladores[i].detencion ? "SI" : "NO"; // bool a string
		std::string status;
		
		switch(controladores[i].estado_detencion) {
			case CONTROLADOR_NORMAL:
			status = "OK";
			break;
			case CONTROLADOR_DETENIENDOSE:
			status = "DETENIENDOSE";
			break;
			case CONTROLADOR_DETENIDO:
			status = "DETENIDO";
			break;
			
		};
	
		std::cout << i + 1 <<"\t";
		std::cout << controladores[i].id << "\t\t" << detencion << "\t" << controladores[i].fmuestreo;
		std::cout << "\t" << controladores[i].conteo << "\t" << controladores[i].multiplo;
		std::cout << "\t" << alarma <<"\t" << controladores[i].err << "\t" << status; 
		std::cout << "\n";
	}	      
	
	std::cout << "\n---------------------------------------------------------------------";
	
	return;
}

template <typename T>
void Watchdog<T>::propiedades_base() { //Propiedades de clase base. Se indica la misma que Dispositivo
	
	Dispositivo::propiedades();
	
	return;
}

template class Watchdog<float>; // Plantilla que se usará. Se ha de declarar aquí, de lo contrario hay error de linker
     



